from django.db.models import TextChoices
from django.utils.translation import gettext_lazy as _


class Status(TextChoices):
    Authorised = "Authorised", _("Authorised")
    Cancel = "Cancel", _("Cancel")
    Created = "Created", _("Created")
    Error = "Error", _("Error")
    Pending = "Pending", _("Pending")
    Refused = "Refused", _("Refused")


class AdyenComponentSupport(TextChoices):
    """
    Adyen-web component support options
    """

    Supported = "supported", _("Supported")
    Unsupported = "unsupported", _("Unsupported")
    Unknown = "unknown", _("Unknown")



import json
import logging
import warnings
from uuid import uuid4

from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from djadyen.constants import ADYEN_WEB_COMPONENTS_SUPPORT
from djadyen.settings import get_setting

from .choices import AdyenComponentSupport, Status

logger = logging.getLogger("adyen")


class AdyenNotification(models.Model):
    created_at = models.DateTimeField(auto_now_add=True)
    notification = models.TextField()
    is_processed = models.BooleanField(default=False)
    processed_at = models.DateTimeField(blank=True, null=True)

    def __str__(self):
        return self.created_at.strftime("%c")

    def get_notification_data(self):
        return json.loads(self.notification)

    def has_status(self, status):
        data = self.get_notification_data()
        event_code = data.get("eventCode")
        merchant_account_code = data.get("merchantAccountCode")
        return event_code == status and merchant_account_code == get_setting(
            "DJADYEN_MERCHANT_ACCOUNT"
        )

    def is_authorised(self, require_success=True):
        data = self.get_notification_data()
        success = data.get("success") == "true"
        has_status = self.has_status("AUTHORISATION")

        if not require_success:
            return has_status

        return has_status and success

    def is_error(self):
        data = self.get_notification_data()
        return self.has_status("ERROR") or data.get("success") == "false"

    def is_cancelled(self):
        return self.has_status("CANCEL")

    def is_refused(self):
        return self.has_status("REFUSED")

    def mark_processed(self, commit=True):
        self.is_processed = True
        self.processed_at = timezone.now()
        if commit:
            self.save()


class AdyenPaymentOption(models.Model):
    name = models.CharField(max_length=200, default="")
    adyen_name = models.CharField(max_length=200, default="")
    guid = models.CharField(max_length=36, verbose_name=_("GUID"), default="")
    image = models.ImageField(null=True)
    is_active = models.BooleanField(default=False)

    def __str__(self):
        return self.name

    def get_adyen_component_support(self):
        if self.adyen_name in ADYEN_WEB_COMPONENTS_SUPPORT:
            return ADYEN_WEB_COMPONENTS_SUPPORT[self.adyen_name]
        return AdyenComponentSupport.Unknown


class AdyenIssuer(models.Model):
    payment_option = models.ForeignKey(AdyenPaymentOption, on_delete=models.CASCADE)
    name = models.CharField(max_length=200, default="")
    adyen_id = models.CharField(max_length=200, default="")

    def __str__(self):
        return self.name


class AdyenOrder(models.Model):
    status = models.CharField(
        max_length=200, choices=Status.choices, default=Status.Created
    )
    status_message = models.TextField(blank=True)
    created_on = models.DateTimeField(auto_now_add=True)
    reference = models.CharField(max_length=200, default="", blank=True)
    psp_reference = models.CharField(max_length=200, default="", blank=True)
    donation_token = models.CharField(
        max_length=1024,
        default="",
        blank=True,
        help_text=_("Donation token to use the 'Giving' web component"),
    )
    email = models.EmailField()

    payment_option = models.ForeignKey(
        AdyenPaymentOption, blank=True, null=True, on_delete=models.SET_NULL
    )
    issuer = models.ForeignKey(
        AdyenIssuer, null=True, blank=True, on_delete=models.SET_NULL
    )

    __old_status = None

    class Meta:
        abstract = True

    def __str__(self):
        return f"{self.reference}"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__old_status = self.status

    def save(self, can_change=False, *args, **kwargs):
        if not self.reference:
            self.reference = uuid4()

        if get_setting("DJADYEN_REFETCH_OLD_STATUS") and self.id:
            self.__old_status = self._meta.model.objects.get(pk=self.id).status

        if self.status != self.__old_status:
            if self.__old_status == Status.Authorised:
                logger.warning(
                    _(
                        "Order ref: %s | Tried to change the "
                        + "status from 'Authorised' to '%s'."
                    ),
                    self.reference,
                    self.status,
                )
                self.status = self.__old_status
            else:
                logger.info(
                    _("Order ref: %s | Changed the status from '%s' to '%s'"),
                    self.reference,
                    self.__old_status,
                    self.status,
                )
                self.__old_status = self.status

        return super().save(*args, **kwargs)

    def process_notification(self, notification):
        if notification.is_authorised():
            self.status = Status.Authorised
            self.save()
            notification.mark_processed()
        elif notification.is_cancelled():
            self.status = Status.Cancel
            self.save()
            notification.mark_processed()
        elif notification.is_refused():
            self.status = Status.Refused
            self.save()
            notification.mark_processed()
        elif notification.is_error():
            self.status = Status.Error
            self.save()
            notification.mark_processed()

        logger.error("Can't process notification with pk %d", notification.pk)

        # Ignore anything else for now.

    def get_price_in_cents(self):
        """
        :return int Return the price in cents for this order.
        """
        raise NotImplementedError

    def get_return_url(self):
        """
        :return str Return the confirmation url after the payment is completed.
        """
        raise NotImplementedError(
            f"Please override 'get_return_url' on the '{self._meta.object_name}'"
        )

    def get_redirect_url(self):
        """
        :return str Return the redirect url after a redirect payment.
        """

        warnings.warn(
            "Implement the payment redirect used in the advanced checkout on "
            f"the '{self._meta.object_name}'. "
            "By default, it will redirect to get_redirect_url",
            stacklevel=2,
        )
        return self.get_return_url()

    def get_payments_api(self):
        """
        :return str Return the payments api endpoint.
        """
        raise NotImplementedError(
            "Implement the payments api endpoint to use "
            f"the advanced checkout on the '{self._meta.object_name}'"
        )

    def get_payment_details_api(self):
        """
        :return str Return the payments api endpoint.
        """
        raise NotImplementedError(
            "Implement the payment details api endpoint to use the advanced "
            f"checkout on the '{self._meta.object_name}'"
        )


class AdyenDonation(models.Model):
    created_on = models.DateTimeField(auto_now_add=True)

    # order needs to be overridden. Djadyen assumes the
    # Djadyen assumes AdyenOrder's related named is `donation_order`
    order = models.OneToOneField(
        AdyenOrder,
        on_delete=models.CASCADE,
        related_name="donation_order",
    )
    status = models.CharField(
        max_length=200, choices=Status.choices, default=Status.Created
    )
    status_message = models.TextField(blank=True)
    campaign = models.CharField(
        _("Campaign ID"), max_length=512, default="", blank=True
    )
    reference = models.CharField(
        _("Donation Reference"),
        max_length=512,
        default="",
        blank=True,
        help_text=_(
            "Donation reference to use the 'Giving' web component. "
            "Separate from the order"
        ),
    )
    amount = models.IntegerField()

    class Meta:
        abstract = True

    def __str__(self):
        return f"{self.reference}"

    def save(self, can_change=False, *args, **kwargs):
        if not self.reference:
            self.reference = uuid4()

        return super().save(*args, **kwargs)



from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from djadyen.choices import AdyenComponentSupport
from djadyen.models import AdyenIssuer, AdyenNotification, AdyenPaymentOption


@admin.register(AdyenNotification)
class AdyenNotificationAdmin(admin.ModelAdmin):
    list_display = ("created_at", "is_processed", "processed_at")
    list_filter = ("is_processed", "created_at", "processed_at")


class AdyenIssuerInline(admin.TabularInline):
    model = AdyenIssuer
    extra = 1


@admin.register(AdyenPaymentOption)
class PaymentOptionAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "adyen_name",
        "is_active",
        "show_issuers",
        "component_support",
    )
    list_filter = ("is_active",)
    inlines = [AdyenIssuerInline]
    search_fields = ("name", "adyen_name")

    @admin.display(description=_("Issuers/Brands"), empty_value="---")
    def show_issuers(self, obj):
        issuers = obj.adyenissuer_set.all()
        if issuers:
            return ", ".join([issuer.adyen_id for issuer in issuers])
        return None

    @admin.display(description=_("Supports Adyen Web"), boolean=True)
    def component_support(self, obj):
        if obj.get_adyen_component_support() == AdyenComponentSupport.Unknown:
            return None
        return obj.get_adyen_component_support() == AdyenComponentSupport.Supported


@admin.register(AdyenIssuer)
class IssuerAdmin(admin.ModelAdmin):
    list_display = ("name", "payment_option")
    list_filter = ("payment_option",)



import json
import logging

from django import template

from djadyen.choices import Status
from djadyen.models import AdyenOrder
from djadyen.settings import get_setting
from djadyen.utils import setup_adyen_client

register = template.Library()
logger = logging.getLogger("adyen")


def get_common_checkout_context(
    language: str,
    order: AdyenOrder,
    country_code: str,
) -> dict:

    return {
        # Adyen Settings
        "client_key": get_setting("DJADYEN_CLIENT_KEY"),
        "environment": get_setting("DJADYEN_ENVIRONMENT"),
        "language": language,
        "country_code": country_code,
        # Djadyen Adyen Settings
        "redirect_url": order.get_return_url,
        "payment_type": (
            order.payment_option.adyen_name if order.payment_option else ""
        ),
        "issuers": (
            json.dumps(
                list(
                    order.payment_option.adyenissuer_set.all().values(
                        "name", "adyen_id"
                    )
                )
            )
            if order.payment_option
            else []
        ),
        "issuer": order.issuer.adyen_id if order.issuer else "",
        "adyen_styles_json": json.dumps(adyen_styles)
        if (adyen_styles := get_setting("DJADYEN_STYLES"))
        else None,
    }


@register.inclusion_tag("adyen/component.html")
def adyen_payment_component(
    language,
    order,
    merchant_account: str = None,
    country_code: str = None,
):
    """
    Will display a singular payment method.
    """
    logger.info("Start new payment for %s", order.reference)
    ady = setup_adyen_client()

    # Setting request data.
    request = {
        "amount": {
            "value": order.get_price_in_cents(),
            "currency": get_setting("DJADYEN_CURRENCYCODE"),
        },
        "reference": str(order.reference),
        "merchantAccount": merchant_account or get_setting("DJADYEN_MERCHANT_ACCOUNT"),
        "returnUrl": order.get_return_url(),
        "shopperLocale": language,
        "countryCode": (
            country_code.lower()
            if country_code
            else get_setting("DJADYEN_DEFAULT_COUNTRY_CODE")
        ),
    }
    try:
        request["shopperEmail"] = order.email
    except Exception:
        pass

    logger.info(request)
    # Starting the checkout.
    result = ady.checkout.payments_api.sessions(
        request, idempotency_key=order.reference
    )

    if result.status_code == 201:
        return {
            "session_id": result.message.get("id"),
            "session_data": result.message.get("sessionData"),
        } | get_common_checkout_context(
            language,
            order,
            request["countryCode"],
        )

    return {}


@register.inclusion_tag("adyen/advanced_component.html")
def adyen_advanced_payment_component(
    language: str,
    order: AdyenOrder,
    country_code: str = None,
):
    return {
        "amount": order.get_price_in_cents(),
        "currency": get_setting("DJADYEN_CURRENCYCODE"),
        "payments_api": order.get_payments_api(),
        "payment_details_api": order.get_payment_details_api(),
    } | get_common_checkout_context(
        language,
        order,
        country_code or get_setting("DJADYEN_DEFAULT_COUNTRY_CODE"),
    )


@register.inclusion_tag("adyen/polling.html")
def adyen_status_polling(order, status_url):
    return {
        "status_url": status_url,
        "pending": order.status in [Status.Created, Status.Pending],
    }


@register.inclusion_tag("adyen/donation_component.html")
def adyen_donation_component(
    language: str,
    campaign: dict,
    redirect_url: str,
    country_code: str = None,
) -> dict:
    """
    Renders the Adyen Giving donation component.
    :param language: Locale of the adyen donation component
    :param campaign: Adyen donation campaign
    :param redirect_url: Redirect url after canceling the donation.
    :param country_code: Adyen Country Code
    :return: Template tag context
    """

    if not campaign:
        return {}

    return {
        "campaign": json.dumps(campaign),
        "campaign_id": campaign["id"],
        "client_key": get_setting("DJADYEN_CLIENT_KEY"),
        "environment": get_setting("DJADYEN_ENVIRONMENT"),
        "language": language,
        "redirect_url": redirect_url,
        "country_code": (
            country_code.lower()
            if country_code
            else get_setting("DJADYEN_DEFAULT_COUNTRY_CODE")
        ),
    }



import json
import logging

try:
    # python 3.13+
    from warnings import deprecated
except ImportError:
    from typing_extensions import deprecated

from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponseRedirect, JsonResponse
from django.shortcuts import redirect
from django.views.generic.detail import DetailView

import Adyen
from Adyen.exceptions import AdyenAPIResponseError
from glom import PathAccessError, glom

from djadyen.choices import Status
from djadyen.constants import LIVE_URL_PREFIX_ERROR
from djadyen.models import AdyenDonation
from djadyen.settings import get_setting
from djadyen.utils import setup_adyen_client

logger = logging.getLogger("adyen")


class AdyenDetailView(DetailView):
    """
    Generic Adyen Order Detail view
    """

    slug_field = "reference"
    slug_url_kwarg = "reference"


class CommonAdyenPaymentView(AdyenDetailView):
    """
    Common functionality for Adyen checkout and payments
    """

    def get_locale(self) -> str:
        """
        Get the Adyen locale. By default, takes the request language code.
        Adyen expects a full locale e.e en-US
        """
        return self.request.LANGUAGE_CODE

    def redirect_ideal_2(self, request):
        logger.info("Start new payment for  %s", self.object.reference)
        ady = setup_adyen_client()

        body = {
            "amount": {
                "value": self.object.get_price_in_cents(),
                "currency": get_setting("DJADYEN_CURRENCYCODE"),
            },
            "paymentMethod": {"type": "ideal"},
            "reference": str(self.object.reference),
            "merchantAccount": get_setting("DJADYEN_MERCHANT_ACCOUNT"),
            "returnUrl": self.object.get_redirect_url(),
            "shopperLocale": self.get_locale(),
            "countryCode": (get_setting("DJADYEN_DEFAULT_COUNTRY_CODE").lower()),
        }

        try:
            request["shopperEmail"] = self.object.email
        except Exception:
            pass

        result = ady.checkout.payments_api.payments(
            body, idempotency_key=f"{self.object.reference}-payments"
        )
        logger.info(request)

        if result.status_code == 200:
            if redirect_url := glom(
                result.message, "action.url", skip_exc=PathAccessError
            ):
                return redirect(redirect_url)

        return redirect(self.object.get_return_url())

    def get_context_data(self, **kwargs) -> dict[str, any]:
        # needs to use adyen web supported language codes e.g. en-US, nl-NL
        return super().get_context_data(**kwargs) | {
            "adyen_language": self.get_locale(),
        }


class AdyenSessionPaymentView(CommonAdyenPaymentView):
    """
    A view which initiates the Adyen payment by rendering a widget.
    It will automatically draw the wanted widget.
    """

    template_name = "adyen/pay.html"

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()

        if self.object.get_price_in_cents() == 0:
            return HttpResponseRedirect(self.object.get_return_url())

        # Ideal 2.0 flow, ideal sorts out the bank itself.
        if (
            self.object.payment_option
            and self.object.payment_option.adyen_name == "ideal"
            and not self.object.issuer
        ):
            return self.redirect_ideal_2(request)
        return super().get(request, *args, **kwargs)


@deprecated(
    "Renamed to `AdyenSessionPaymentView` to differentiate "
    "from the Advanced payment view."
)
class AdyenPaymentView(AdyenSessionPaymentView): ...


class AdyenResponseView(AdyenDetailView):
    def get(self, request, *args, **kwargs):
        self.object = self.get_object()

        if self.object.get_price_in_cents() == 0:
            self.handle_authorised()

        resultRedirect = request.GET.get("redirectResult")
        if resultRedirect:
            ady = Adyen.Adyen()

            if not get_setting("DJADYEN_ENVIRONMENT"):
                raise ImproperlyConfigured("Please provide an environment.")

            if get_setting("DJADYEN_ENVIRONMENT") == "live" and not get_setting(
                "DJADYEN_LIVE_URL_PREFIX"
            ):
                raise ImproperlyConfigured(LIVE_URL_PREFIX_ERROR)

            # Setting global values
            ady.payment.client.platform = get_setting("DJADYEN_ENVIRONMENT")
            ady.payment.client.xapikey = get_setting("DJADYEN_SERVER_KEY")
            ady.payment.client.app_name = get_setting("DJADYEN_APPNAME")
            ady.payment.client.live_endpoint_prefix = get_setting(
                "DJADYEN_LIVE_URL_PREFIX"
            )

            # Setting request data.
            request = {
                "details": {
                    "redirectResult": resultRedirect,
                },
            }
            # Requesting the status.
            result = ady.checkout.payments_api.payments_details(request)
            result_code = result.message.get("resultCode")
            if result_code == "Authorised":
                self.handle_authorised()
            elif result_code != "Pending":
                self.handle_error()
        return super().get(request, *args, **kwargs)

    def handle_authorised(self) -> None:
        """
        Handle what happens to a order after a payment is authorised
        e.g change order to status and send confirmation email
        """
        raise NotImplementedError(
            "Handle what happens to a order after a payment is authorised. "
            "Use self.object"
        )

    def handle_error(self):
        self.object.status = Status.Error
        self.object.save()


class AdyenStatusView(DetailView):
    slug_field = "reference"
    slug_url_kwarg = "reference"

    def get(self, request, *args, **kwargs):
        order = self.get_object()
        updated_status = True
        if order.status in [Status.Created, Status.Pending]:
            updated_status = False

        return JsonResponse(
            {
                "updatedStatus": updated_status,
                "currentStatus": order.status,
                "reference": order.reference,
            }
        )


@deprecated("Renamed to `AdyenStatusView` as it supports both orders and donations.")
class AdyenOrderStatusView(AdyenStatusView): ...


class AdyenAdvancedPaymentView(CommonAdyenPaymentView):
    template_name = "adyen/advanced_pay.html"

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()

        if self.object.get_price_in_cents() == 0:
            return HttpResponseRedirect(self.object.get_return_url())

        # Ideal 2.0 flow, ideal sorts out the bank itself.
        # Skip if redirectResult is present, as it means the payment is already done
        # and can cause loops.
        if (
            not request.GET.get("redirectResult")
            and self.object.payment_option
            and self.object.payment_option.adyen_name == "ideal"
            and not self.object.issuer
        ):
            return self.redirect_ideal_2(request)

        return super().get(request, *args, **kwargs)


class AdyenDonationView(CommonAdyenPaymentView):
    """
    Allows dontaiton option after finishing the original payment.
    Should be done
    """

    template_name = "adyen/donation.html"

    def get_donation(self):
        ady = setup_adyen_client()
        json_request = {
            "merchantAccount": get_setting("DJADYEN_MERCHANT_ACCOUNT"),
            "currency": get_setting("DJADYEN_CURRENCYCODE"),
            "locale": self.get_locale(),
        }

        result = ady.checkout.donations_api.donation_campaigns(request=json_request)
        if result.status_code == 200 and result.message.get("donationCampaigns"):
            return result.message.get("donationCampaigns")[0]
        return None

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["campaign"] = self.get_donation()
        context["redirect_url"] = self.get_redirect_url()
        return context

    def get_redirect_url(self) -> str:
        """
        Where to redirect after closing the donation component
        :return: URL
        """
        return "/"

    def get_donation_model(self) -> AdyenDonation:
        raise NotImplementedError("Donation model is required to create a donation")

    def get_donation_confirmation_url(self) -> str:
        raise NotImplementedError(
            "Donation confirmation url is required to redirect to after donation"
        )

    def handle_authorised(self) -> None:
        """
        Handle what happens to a order after a payment is authorised
        e.g change order to status and send confirmation email
        """
        raise NotImplementedError(
            "Handle what happens to a order after a payment is authorised. "
            "Use self.object"
        )

    def get(self, request, *args, **kwargs):
        self.object = self.get_object()

        # handle free payment
        if self.object.get_price_in_cents() == 0:
            self.handle_authorised()

        return super().get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        self.object = self.get_object()
        ady = setup_adyen_client()

        amount = json.loads(request.POST["amount"])
        amount_value = amount["value"]
        campaign_id = request.POST["donation_campaign_id"]

        logger.info("Start new donation payment for  %s", self.object.reference)

        Donation = self.get_donation_model()
        donation, created = Donation.objects.get_or_create(
            order=self.object,
            defaults={"amount": amount_value, "campaign": campaign_id},
        )

        # if the donation API is not sent yet
        if donation.status == Status.Created.value:
            # if donation is created but not run, donation can be overrided
            if not created:
                donation.amount = amount_value
                donation.campaign = campaign_id

            try:
                # iDeal is sepadirectdebit
                # other current supported payment methods are scheme
                payemnt_type: str = (
                    "sepadirectdebit"
                    if self.object.payment_option.adyen_name == "ideal"
                    else "scheme"
                )

                donation_request = {
                    "amount": amount,
                    "donationCampaignId": campaign_id,
                    "paymentMethod": {"type": payemnt_type},
                    "donationOriginalPspReference": self.object.psp_reference,
                    "donationToken": self.object.donation_token,
                    "reference": str(donation.reference),
                    "merchantAccount": get_setting("DJADYEN_MERCHANT_ACCOUNT"),
                }
                result = ady.checkout.donations_api.donations(
                    donation_request, idempotency_key=str(donation.reference)
                )

            except AdyenAPIResponseError as e:
                logger.error(
                    "Error when creating donaiton %s: %s", donation.reference, e
                )
                donation.status_message = str(e)
                donation.status = Status.Error.value
                donation.save()
            else:
                logger.info("Donation created %s", donation.reference)
                if result.message["status"] == "refused":
                    donation.status = Status.Refused.value
                else:
                    donation.status = Status.Pending.value

                # TODO: better status message?
                donation.status_message = result.message
                donation.save()

        return redirect(self.get_donation_confirmation_url())



import base64
import binascii
import hashlib
import hmac
import logging

from djadyen.settings import get_setting

logger = logging.getLogger("adyen")


def get_signature(params):
    """
    Description of how the HMAC can be signed can be found here

    https://docs.adyen.com/development-resources/webhooks/verify-hmac-signatures
    """
    hmac_key = binascii.a2b_hex(get_setting("DJADYEN_NOTIFICATION_KEY"))

    logger.debug("Params: %s", params)

    signing_string = "{part1}:{part2}".format(
        part1="{psp_reference}:{original_reference}:{merchant_account_code}".format(
            psp_reference=params.get("pspReference", ""),
            original_reference=params.get("originalReference", ""),
            merchant_account_code=params.get("merchantAccountCode", ""),
        ),
        part2="{merchant_reference}:{value}:{currency}:{event_code}:{success}".format(
            merchant_reference=params.get("merchantReference", ""),
            value=params.get("amount", {}).get("value", ""),
            currency=params.get("amount", {}).get("currency", ""),
            event_code=params.get("eventCode", ""),
            success=params.get("success", ""),
        ),
    )
    logger.debug("Signing Params: %s", signing_string)

    hmac_string = hmac.new(hmac_key, signing_string.encode("utf-8"), hashlib.sha256)
    logger.debug("HMAC: %s", hmac_string)

    signature = base64.b64encode(hmac_string.digest()).decode("utf-8")
    logger.debug("signature: %s", signature)
    return signature



from datetime import timedelta

from django.apps import apps
from django.core.management.base import BaseCommand
from django.db.models import Q
from django.utils import timezone

from djadyen.choices import Status
from djadyen.models import AdyenNotification
from djadyen.settings import get_setting


class Command(BaseCommand):
    help = "Process the adyen notifications that are not processed yet."

    def handle(self, *args, **options):
        # Add a timeout to not process all the notifications
        # after the view is already processed.
        time_ago = timezone.now() - timedelta(
            minutes=get_setting("DJADYEN_HANDLE_NOTIFICATION_MINUTES_AGO")
        )

        order_models = [
            apps.get_model(model) for model in get_setting("DJADYEN_ORDER_MODELS")
        ]

        # Process notifications which have been sent by Adyen.
        for notification in AdyenNotification.objects.filter(
            is_processed=False, created_at__lte=time_ago
        ):
            notification_data = notification.get_notification_data()
            reference = notification_data.get("merchantReference")
            for order_model in order_models:
                #
                # TOOD: Ugh, okay so we process only 'Pending' orders, this might
                # or might not be correct, I don't understand the adyen state machine
                # well enough.
                #
                # The reason this is done to avoid that an order, which is already
                # processed in the return URL, to be processed again by a notification
                #
                orders = order_model.objects.filter(reference=reference).filter(
                    Q(status=Status.Pending) | Q(status=Status.Created)
                )

                for order in orders:
                    order.process_notification(notification)

        # After five days of an Order having status 'Pending', move them to 'Error'
        five_days_ago = timezone.now() - timedelta(days=5)
        for order_model in order_models:
            for obj in order_model.objects.filter(
                status=Status.Pending, created_on__lte=five_days_ago
            ):
                obj.status = Status.Error
                obj.save()

        # Close 5 day old notifications
        for notification in AdyenNotification.objects.filter(
            is_processed=False, created_at__lte=five_days_ago
        ):
            notification.mark_processed()



from djadyen.choices import AdyenComponentSupport, Status

LIVE_URL_PREFIX_ERROR = "Please provide the live_url_prefix. "
"https://docs.adyen.com/development-resources/live-endpoints#live-url-prefix"

"""
If an adyen name is supported or not by the web components.
"""
ADYEN_WEB_COMPONENTS_SUPPORT = {
    "alipay": AdyenComponentSupport.Supported,
    "bankTransfer_IBAN": AdyenComponentSupport.Supported,  # SEPA Bank Transfer
    "bcmc": AdyenComponentSupport.Supported,  # Bancontact card
    "directEbanking": AdyenComponentSupport.Unsupported,  # Online bank transfer - only API # noqa: E501
    "ebanking_FI": AdyenComponentSupport.Supported,  # Finnish E-Banking
    "ideal": AdyenComponentSupport.Supported,
    "scheme": AdyenComponentSupport.Supported,  # Credit/Debit cards
    "sepadirectdebit": AdyenComponentSupport.Supported,  # SEPA Direct Debit
}

"""
'These result codes indicate that the payment has reached a final state.'
https://docs.adyen.com/online-payments/build-your-integration/payment-result-codes#final-payment-status
"""
ADYEN_FINAL_STATE_CODES = (
    Status.Authorised.value,
    Status.Cancel.value,
    Status.Error.value,
    Status.Refused.value,
)



import json
import logging

from django.http import JsonResponse
from django.views import View
from django.views.generic.detail import SingleObjectMixin

from Adyen.exceptions import AdyenAPIResponseError

from djadyen.choices import Status
from djadyen.constants import ADYEN_FINAL_STATE_CODES
from djadyen.models import AdyenOrder
from djadyen.settings import get_setting
from djadyen.utils import setup_adyen_client

logger = logging.getLogger("adyen")


def handle_adyen_error_response(
    api_name: str, obj: AdyenOrder, exception: AdyenAPIResponseError
) -> None:
    if exception.psp:
        obj.psp_reference = exception.psp

    if obj.status != Status.Authorised:
        obj.status = Status.Error
    obj.status_message = f"Adyen API {api_name} - {exception}"
    obj.save()


class AdyenAPIView(SingleObjectMixin, View):
    slug_field = "reference"
    slug_url_kwarg = "reference"


class AdyenPaymentsAPI(AdyenAPIView):
    def post(self, request, *args, **kwargs):
        self.object = self.get_object()

        try:
            data = json.loads(request.body)
        except json.JSONDecodeError:
            return JsonResponse({"error": "bad response"}, status=400)

        payment_method = data.get("paymentMethod")

        if not payment_method:
            return JsonResponse({"error": "bad response"}, status=400)
        json_request = {
            "amount": {
                "currency": "EUR",
                "value": self.object.get_price_in_cents(),
            },
            "reference": str(self.object.reference),
            "paymentMethod": payment_method,
            "returnUrl": self.object.get_redirect_url(),
            "merchantAccount": get_setting("DJADYEN_MERCHANT_ACCOUNT"),
        }

        if data.get("riskData"):
            json_request["riskData"] = data["riskData"]

        if data.get("checkoutAttemptId"):
            json_request["checkoutAttemptId"] = data["checkoutAttemptId"]

        # Send the request
        logger.info("Start new payment for  %s", self.object.reference)
        adyen_client = setup_adyen_client()

        try:
            result = adyen_client.checkout.payments_api.payments(
                request=json_request,
                idempotency_key=f"{self.object.reference}-payments",
            )
        except AdyenAPIResponseError as e:
            handle_adyen_error_response("/payments/", self.object, e)
            return JsonResponse({"error": "bad response"}, status=400)

        # only return what checkout wants from payments
        response = {
            "resultCode": result.message["resultCode"],
            "action": result.message.get("action"),
            "order": result.message.get("order"),
            "donationToken": result.message.get("donationToken"),
        }

        self.object.status = Status.Pending.value
        self.object.psp_reference = result.message.get("pspReference", "")

        if response["donationToken"]:
            self.object.donation_token = response["donationToken"]
        self.object.save()

        return JsonResponse(response, status=200)


class AdyenPaymentDetailsAPI(AdyenAPIView):
    def post(self, request, *args, **kwargs):
        self.object = self.get_object()

        try:
            data = json.loads(request.body)
        except json.JSONDecodeError:
            return JsonResponse({"error": "bad response"}, status=400)

        if not data:
            return JsonResponse({"error": "bad response"}, status=400)

        # STATE_DATA is an object passed from your client app,
        # deserialized from JSON to a data structure.

        logger.info("Sending payment details for  %s", self.object.reference)
        adyen_client = setup_adyen_client()

        # Setting request data. For now assumes that /details/ API is always a redirect.
        request = {
            "details": data,
        }
        try:
            result = adyen_client.checkout.payments_api.payments_details(
                request, idempotency_key=f"{self.object.reference}-detail"
            )
        except AdyenAPIResponseError as e:
            handle_adyen_error_response("/payment/details/", self.object, e)
            return JsonResponse({"error": "bad response"}, status=400)

        result_code = result.message["resultCode"]

        # only return what checkout wants from payment details
        response = {
            "resultCode": result_code,
            "action": result.message.get("action"),
            "order": result.message.get("order"),
            "donationToken": result.message.get("donationToken"),
        }

        self.object.status = Status.Pending.value
        if result.message.get("pspReference"):
            self.object.psp_reference = result.message["pspReference"]

        if response["donationToken"]:
            self.object.donation_token = response["donationToken"]
        self.object.save()

        # iDeal can be Authorised within the api response
        if result_code in ADYEN_FINAL_STATE_CODES:
            if result_code == "Authorised":
                self.handle_authorised()
            else:
                self.handle_error(result_code)

        return JsonResponse(response)

    def handle_authorised(self) -> None:
        """
        Handle what happens to an order after a payment is authorized
        e.g. change order to status and send a confirmation email
        """

        raise NotImplementedError(
            "Handle what happens to a order after a payment is authorised"
        )

    def handle_error(self, result_code: str):
        self.object.status = result_code
        self.object.save()



from datetime import datetime

from django.core.management import call_command
from django.test.testcases import TestCase

import requests_mock
from freezegun import freeze_time

from djadyen.choices import Status
from djadyen.models import AdyenIssuer, AdyenPaymentOption

from .factories import NotificationFactory, OrderFactory
from .utils import TestFileMixin


class SyncPaymentMethods(TestFileMixin, TestCase):
    @requests_mock.mock()
    def test_on_empty_database_mock(self, mock):
        mock.post(
            "https://checkout-test.adyen.com/v71/paymentMethods",
            [
                {
                    "content": self._get_test_file("payment_methods.json").read(),
                    "status_code": 200,
                },
            ],
        )

        call_command("sync_payment_methods")

        self.assertEqual(AdyenPaymentOption.objects.count(), 9)
        self.assertEqual(AdyenIssuer.objects.count(), 57)

    @requests_mock.mock()
    def test_on_existing_database_mock(self, mock):
        mock.post(
            "https://checkout-test.adyen.com/v71/paymentMethods",
            [
                {
                    "content": self._get_test_file("payment_methods.json").read(),
                    "status_code": 200,
                },
                {
                    "content": self._get_test_file("payment_methods.json").read(),
                    "status_code": 200,
                },
            ],
        )
        self.assertEqual(AdyenPaymentOption.objects.count(), 0)
        self.assertEqual(AdyenIssuer.objects.count(), 0)

        call_command("sync_payment_methods")

        self.assertEqual(AdyenPaymentOption.objects.count(), 9)
        self.assertEqual(AdyenIssuer.objects.count(), 57)

        call_command("sync_payment_methods")

        self.assertEqual(AdyenPaymentOption.objects.count(), 9)
        self.assertEqual(AdyenIssuer.objects.count(), 57)

    @requests_mock.mock()
    def test_sync_brands(self, mock):
        """
        Test payment brands converted to issuer objects.
        """

        mock.post(
            "https://checkout-test.adyen.com/v71/paymentMethods",
            [
                {
                    "content": self._get_test_file("payment_methods.json").read(),
                    "status_code": 200,
                },
            ],
        )

        call_command("sync_payment_methods")

        bancontact = AdyenPaymentOption.objects.get(adyen_name="bcmc")
        self.assertEqual(bancontact.adyenissuer_set.all().count(), 3)

        card = AdyenPaymentOption.objects.get(adyen_name="scheme")
        self.assertEqual(card.adyenissuer_set.all().count(), 7)


class ProcessNotifications(TestFileMixin, TestCase):
    def setUp(self):
        super().setUp()

        reference = "93ca495f-3b85-4df7-aeb0-052194014c2e"

        with freeze_time("2019-01-01 11:44"):
            self.notification1 = NotificationFactory.create(
                notification=self._get_json_data("notification_data_success.json"),
                is_processed=False,
            )
        self.order1 = OrderFactory.create(status=Status.Pending, reference=reference)

    @freeze_time("2019-01-01 12:00")
    def test_process_notifications_already_processed(self):
        """
        Make sure that an order, which status has already been
        set as 'Authorised' is not processed again.
        """
        self.order1.status = Status.Authorised
        self.order1.save()

        self.assertFalse(self.order1.paid)

        call_command("adyen_maintenance")

        self.order1.refresh_from_db()
        self.assertFalse(self.order1.paid)

    @freeze_time("2019-01-01 12:00")
    def test_process_notifications(self):
        self.assertFalse(self.order1.paid)

        call_command("adyen_maintenance")

        self.order1.refresh_from_db()
        self.assertTrue(self.order1.paid)

        self.notification1.refresh_from_db()
        self.assertTrue(self.notification1.is_processed)
        self.assertTrue(self.notification1.processed_at, datetime(2019, 1, 1, 12, 0))

    @freeze_time("2019-01-01 12:00")
    def test_process_notifications_unsuccessful(self):
        self.assertFalse(self.order1.paid)

        self.notification1.notification = self._get_json_data(
            "notification_data_unsuccessful.json"
        )
        self.notification1.save()

        call_command("adyen_maintenance")

        self.order1.refresh_from_db()
        self.assertFalse(self.order1.paid)

        self.notification1.refresh_from_db()
        self.assertTrue(self.notification1.is_processed)
        self.assertTrue(self.notification1.processed_at, datetime(2019, 1, 1, 12, 0))


class CleanupPending(TestFileMixin, TestCase):
    def test_cleanup(self):
        # 5 days ago; Should be marked as 'Error'
        with freeze_time("2019-01-5 12:00"):
            self.order1 = OrderFactory.create(status=Status.Pending)
        # 4 days ago; Should be left alone
        with freeze_time("2019-01-6 12:00"):
            self.order2 = OrderFactory.create(status=Status.Pending)
        # 6 days ago, Should be marked as 'Error'
        with freeze_time("2019-01-4 12:00"):
            self.order3 = OrderFactory.create(status=Status.Pending)
        # 7 days ago, but Authorised, should be left alone
        with freeze_time("2019-01-3 12:00"):
            self.order4 = OrderFactory.create(status=Status.Authorised)

        with freeze_time("2019-01-01 11:44"):
            self.notification1 = NotificationFactory.create(
                notification=self._get_json_data(
                    "notification_data_reference_unknown.json"
                ),
                is_processed=False,
            )

        with freeze_time("2019-01-10 12:00"):
            call_command("adyen_maintenance")

        self.order1.refresh_from_db()
        self.order2.refresh_from_db()
        self.order3.refresh_from_db()
        self.order4.refresh_from_db()
        self.notification1.refresh_from_db()

        self.assertEqual(self.order1.status, Status.Error)
        self.assertEqual(self.order2.status, Status.Pending)
        self.assertEqual(self.order3.status, Status.Error)
        self.assertEqual(self.order4.status, Status.Authorised)
        self.assertTrue(self.notification1.is_processed)



from typing import Literal

from django.conf import settings

DJADYEN_SERVER_KEY: str = ""
"""
The Adyen API key used in backend API requests. Required
https://docs.adyen.com/development-resources/api-authentication#api-key-authentication
"""

DJADYEN_CLIENT_KEY: str = ""
"""
The Adyen environment to used in frontend API requests. Required.
https://docs.adyen.com/development-resources/client-side-authentication
"""

DJADYEN_ENVIRONMENT: str = "test"
"""
The Adyen environment to use. Defaults to `test`
"""

DJADYEN_MERCHANT_ACCOUNT: str = ""
"""
Merchant account from Adyen
"""


DJADYEN_APPNAME: str = "Djadyen Payment"
"""
Django's app name for Adyen. Defaults to `Djadyen Payment`
"""


DJADYEN_CURRENCYCODE: str = "EUR"
"""
Adyen current code used in payments. Defaults to `EUR`
https://docs.adyen.com/development-resources/currency-codes#currency-codes
"""


DJADYEN_REFETCH_OLD_STATUS: bool = False
"""
Fetch the older status when updating an order.

Defaults to `False`.
"""


DJADYEN_ORDER_MODELS: list["str"] = []
"""
List of DjAdyen orders that should be handled by the webhook.
e.g. ["app.Order", "app.Donation"]
"""


DJADYEN_HANDLE_NOTIFICATION_MINUTES_AGO: int = 0
"""
Delay to handle notifications in minutes. e.g 0 is immediate, 1 is older than 1 minute.
Default is 0.
"""

DJADYEN_NOTIFICATION_KEY: str = ""
"""
HMAC signatures to verify notifications are sent by Adyen.
https://docs.adyen.com/development-resources/webhooks/verify-hmac-signatures
"""


DJADYEN_DEFAULT_COUNTRY_CODE: str = ""
"""
Default shopper's two-letter country code.
"""

DJADYEN_LIVE_URL_PREFIX: str = ""
"""
The live URL prefix from adyen required to use live payments.
https://docs.adyen.com/development-resources/live-endpoints#live-url-prefix

Defaults to empty string. Required to not be empty for live
"""


DJADYEN_STYLES: dict = {}
"""
Styles are passed to the Adyen payment component configuration.
Defaults to `{}`

Example setting:
DJADYEN_STYLES = {
    'base': {
        'color': '#000000',
        'fontSize': '16px',
        'fontFamily': 'Arial, sans-serif',
    },
    'placeholder': {
        'color': '#999999',
    },
    'error': {
        'color': '#ff0000',
    }
}
"""


# TODO: type statement after 3.11 support is dropped
SettingName = Literal[
    "DJADYEN_SERVER_KEY",
    "DJADYEN_CLIENT_KEY",
    "DJADYEN_ENVIRONMENT",
    "DJADYEN_APPNAME",
    "DJADYEN_MERCHANT_ACCOUNT",
    "DJADYEN_CURRENCYCODE",
    "DJADYEN_REFETCH_OLD_STATUS",
    "DJADYEN_HANDLE_NOTIFICATION_MINUTES_AGO",
    "DJADYEN_ORDER_MODELS",
    "DJADYEN_NOTIFICATION_KEY",
    "DJADYEN_DEFAULT_COUNTRY_CODE",
    "DJADYEN_LIVE_URL_PREFIX",
    "DJADYEN_STYLES",
]


def get_setting(name: SettingName):
    default = globals()[name]
    return getattr(settings, name, default)



import logging

from django.core.exceptions import ImproperlyConfigured

import Adyen

from djadyen.constants import LIVE_URL_PREFIX_ERROR
from djadyen.settings import get_setting

logger = logging.getLogger("adyen")


def setup_adyen_client():
    ady = Adyen.Adyen()

    if not get_setting("DJADYEN_ENVIRONMENT"):
        raise ImproperlyConfigured("Please provide an environment.")

    if get_setting("DJADYEN_ENVIRONMENT") == "live" and not get_setting(
        "DJADYEN_LIVE_URL_PREFIX"
    ):
        raise ImproperlyConfigured(LIVE_URL_PREFIX_ERROR)

    # Setting global values
    ady.payment.client.platform = get_setting("DJADYEN_ENVIRONMENT")
    ady.payment.client.xapikey = get_setting("DJADYEN_SERVER_KEY")
    ady.payment.client.app_name = get_setting("DJADYEN_APPNAME")
    ady.payment.client.live_endpoint_prefix = get_setting("DJADYEN_LIVE_URL_PREFIX")

    return ady



from django.test import TestCase, override_settings

import pytest

from djadyen.choices import Status

from .factories import (
    IssuerFactory,
    NotificationFactory,
    OrderFactory,
    PaymentOptionsFactory,
)


def test_order_model_required_implements():
    from testapp.models import BadOrderModel

    order = BadOrderModel()

    with pytest.raises(NotImplementedError):
        order.get_return_url()

    order.get_return_url = lambda: "https://example.com/get_return_url"

    with pytest.warns(UserWarning) as record:
        url = order.get_redirect_url()
        assert url == "https://example.com/get_return_url"

    assert len(record) == 1
    assert (
        record[0].message.args[0]
        == "Implement the payment redirect used in the advanced checkout on the "
        "'BadOrderModel'. By default, it will redirect to get_redirect_url"
    )


class AdyenNofiticationTests(TestCase):
    def test_str(self):
        notification = NotificationFactory()
        self.assertEqual(notification.__str__(), notification.created_at.strftime("%c"))


class AdyenOrderTests(TestCase):
    def test_str(self):
        with self.assertNumQueries(1):
            order = OrderFactory()
            self.assertEqual(order.__str__(), f"{order.reference}")

    @override_settings(ADYEN_REFETCH_OLD_STATUS=True)
    def test_refetch_old_status(self):
        with self.assertNumQueries(2):
            order = OrderFactory()
            self.assertEqual(order.__str__(), f"{order.reference}")

            order.status = Status.Authorised
            order.save()
            self.assertEqual(order.__str__(), f"{order.reference}")

    def test_can_not_overwrite_authorised_status(self):
        with self.assertNumQueries(2):
            order = OrderFactory(status=Status.Authorised)
            self.assertEqual(order.status, "Authorised")

            order.status = Status.Cancel
            order.save()
            self.assertEqual(order.status, "Authorised")


class AdyenIssuerTests(TestCase):
    def test_str(self):
        issuer = IssuerFactory()
        self.assertEqual(issuer.__str__(), issuer.name)


class AdyenPaymentOptionTests(TestCase):
    def test_str(self):
        payment_option = PaymentOptionsFactory()
        self.assertEqual(payment_option.__str__(), payment_option.name)
